@@ -3,7 +3,9 @@
 #  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
 #  ------------------------------------------------------------------------------------------
 
-from health.azure.azure_util import fetch_run, set_environment_variables_for_multi_node, split_recovery_id
+from health.azure.azure_util import (fetch_run, set_environment_variables_for_multi_node, split_recovery_id,
+                                     get_most_recent_run, download_files_from_run_id, download_checkpoints_from_run_id,
+                                     download_from_datastore, upload_to_datastore, torch_barrier)
 from health.azure.datasets import DatasetConfig
 from health.azure.himl import (AzureRunInfo, create_run_configuration, create_script_run, get_workspace, submit_run,
                                submit_to_azure_if_needed)
@@ -17,4 +19,10 @@ __all__ = ["fetch_run",
            "create_script_run",
            "get_workspace",
            "submit_run",
-           "submit_to_azure_if_needed"]
+           "submit_to_azure_if_needed",
+           "get_most_recent_run",
+           "download_files_from_run_id",
+           "download_checkpoints_from_run_id",
+           "download_from_datastore",
+           "upload_to_datastore",
+           "torch_barrier"]
